/*
 * sos/sos.c --
 *
 *	There should be a compiler appreciation day. This file
 *	demonstrates why. Compile with and without optimization,
 *	look at the generated assembly code, play with it (e.g.,
 *	remove the static keywords, what does change?).
 *
 *      Here is the Haskell version of sos:
 *
 *      sos :: Integer -> Integer -> Integer
 *      sos a b = sum [i^2 | i <- [a..b]]
 *
 *      See the corresponding Haskell file for additional Haskell
 *      versions.
 *
 *      Try:
 *      ./sos 0 4
 *      ./sos 0 4444
 *      ./sos 0 4444444
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#if 1
#define static
#endif

static int64_t sos(int64_t a, int64_t b)
{
    int64_t i, s = 0, n = b-a+1;
    int64_t *p;
    
    if (n < 0) {
	return 0;
    }
    
    p = (int64_t *) malloc(n * sizeof(int64_t));
    if (! p) {
	fprintf(stderr, "malloc failure\n");
	exit(EXIT_FAILURE);
    }
    
    for (i = 0; i < n; i++) {
	p[i] = a++;
    }
    
    for (i = 0; i < n; i++) {
	p[i] = p[i] * p[i];
    }
    
    for (i = 0; i < n; i++) {
	s += p[i];
    }
    
    free(p);
    return s;
}

static int64_t sos_(int64_t a, int64_t b)
{
    return (a > b) ? 0 : a*a + sos_(a+1, b);
}

static int64_t sosh(int64_t a, int64_t b, int64_t c)
{
    return (a > b) ? c : sosh(a+1, b, c+a*a);
}

static int64_t sos__(int64_t a, int64_t b)
{
    return sosh(a, b, 0);
}

static int64_t sos___(int64_t a, int64_t b)
{
    int64_t s = 0;
    for (s = 0; a <= b; a++) {
	s += a*a;
    }
    return s;
}

int main(int argc, char *argv[])
{
    int64_t a, b;
    
    if (argc != 3) {
	fprintf(stderr, "usage: sos min max\n");
	return EXIT_FAILURE;
    }

    a = strtoll(argv[1], NULL, 10);
    b = strtoll(argv[2], NULL, 10);
    
    printf("sos %lld %lld = %lld\n", a, b, sos(a, b));
    printf("sos_ %lld %lld = %lld\n", a, b, sos_(a, b));
    printf("sos__ %lld %lld = %lld\n", a, b, sos__(a, b));
    printf("sos___ %lld %lld = %lld\n", a, b, sos___(a, b));
    
    return EXIT_SUCCESS;
}
